__version__ = "0.3.1"
from .remapper import apply, apply_lr, get_map
from .transformer import (
    DenormalizeTransformer,
    EquirectangularEncoder,
    Euclidean3DRotator,
    Euclidean3DTransformer,
    FisheyeDecoder,
    FisheyeEncoder,
    MultiTransformer,
    NormalizeTransformer,
    PolarRollTransformer,
    TransformerBase,
    ZoomTransformer,
)

__all__ = [
    "TransformerBase",
    "ZoomTransformer",
    "MultiTransformer",
    "NormalizeTransformer",
    "PolarRollTransformer",
    "DenormalizeTransformer",
    "FisheyeDecoder",
    "FisheyeEncoder",
    "EquirectangularEncoder",
    "Euclidean3DRotator",
    "Euclidean3DTransformer",
    "apply",
    "apply_lr",
    "get_map",
]



from __future__ import annotations

from colorsys import hls_to_rgb
from pathlib import Path

import cv2 as cv
import numpy as np
from numpy.typing import NDArray


def generate_test_image(
    size: int = 2048, path: str | Path | None = None
) -> NDArray[np.uint8]:
    """
    Generate a test image.

    Parameters
    ----------
    size : int, optional
        The height and width of the image, by default 2048
    path : str | Path | None, optional
        The path to save the image, by default None
        If None, the image will not be saved.

    Returns
    -------
    NDArray[np.uint8]
        The generated image.

    """
    img = np.zeros((size, size, 3), dtype=np.uint8)
    center = size // 2
    scale = size // 512 + 1
    # draw circle
    for radius in np.linspace(0, center, 10, endpoint=True):
        color = hls_to_rgb(radius / center, 0.5, 1)
        color = tuple(int(c * 255) for c in color)  # type: ignore
        cv.circle(img, (center, center), int(radius), color, scale)
        for angle in np.linspace(0, np.pi * 2, 4, endpoint=False):
            cv.putText(
                img,
                f"{radius / center * 90:g}",
                (
                    int(center + np.cos(angle) * radius),
                    int(center + np.sin(angle) * radius),
                ),
                cv.FONT_HERSHEY_SIMPLEX,
                scale // 2,
                color,
                scale // 2,
            )
    # draw line
    for angle in np.linspace(0, np.pi * 2, 24, endpoint=False):
        color = hls_to_rgb(angle / (np.pi * 2), 0.5, 1)
        color = tuple(int(c * 255) for c in color)  # type: ignore
        x = center + np.cos(angle) * center
        y = center + np.sin(angle) * center
        cv.line(img, (center, center), (int(x), int(y)), color, scale)
    if path:
        cv.imwrite(Path(path).as_posix(), img)
    return img



from __future__ import annotations

import warnings
from abc import ABCMeta, abstractmethod
from typing import Any, Generic, Literal, Sequence, TypeVar

import attrs
import numpy as np
from numpy.typing import NDArray
from quaternion import quaternion, rotate_vectors
from sklearn.base import BaseEstimator, TransformerMixin


class TransformerBase(
    BaseEstimator,
    TransformerMixin,
    metaclass=ABCMeta,
):
    """Base class for transformers."""

    # def fit(self, image: NDArray, **kwargs: Any) -> None:
    #     pass

    @abstractmethod
    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        """
        Transform the input coordinates.

        Parameters
        ----------
        x : NDArray
            x (left-right) coordinates.
        y : NDArray
            y (up-down) coordinates.
        **kwargs : Any
            Any additional keyword arguments.

        Returns
        -------
        tuple[NDArray, NDArray]
            x and y coordinates after transformation.

        """
        pass

    @abstractmethod
    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        """
        Inverse transform the input coordinates.

        Parameters
        ----------
        x : NDArray
            x (left-right) coordinates.
        y : NDArray
            y (up-down) coordinates.
        **kwargs : Any
            Any additional keyword arguments.

        Returns
        -------
        tuple[NDArray, NDArray]
            x and y coordinates after transformation.

        """

    def __mul__(self, other: TransformerBase) -> MultiTransformer:
        """Multiply two transformers together."""
        if isinstance(self, MultiTransformer) and isinstance(other, MultiTransformer):
            return MultiTransformer(
                transformers=[*self.transformers, *other.transformers]
            )
        if isinstance(self, MultiTransformer):
            return MultiTransformer(transformers=[*self.transformers, other])
        if isinstance(other, MultiTransformer):
            return MultiTransformer(transformers=[self, *other.transformers])
        return MultiTransformer(transformers=[self, other])


T = TypeVar("T", bound=TransformerBase)


@attrs.define()
class MultiTransformer(TransformerBase):
    """A transformer that applies multiple transformers in sequence."""

    transformers: list[TransformerBase]

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        for transformer in self.transformers:
            x, y = transformer.transform(x, y, **kwargs)
        return x, y

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        for transformer in reversed(self.transformers):
            x, y = transformer.inverse_transform(x, y, **kwargs)
        return x, y


def get_radius(input: NDArray, *, threshold: int = 10) -> float:
    """
    Estimate the radius of the circle in the image.

    Parameters
    ----------
    input : NDArray
        The input image.
    threshold : int, optional
        The threshold to determine if a pixel is black, by default 10

    Returns
    -------
    float
        The estimated radius.

    """
    height, width = input.shape[:2]
    if width > height:
        center_row = input[height // 2, :, :]
    else:
        center_row = input[:, width // 2, :]
    del height, width

    # determine if a pixel is black
    center_row_is_black = np.mean(center_row, axis=-1) < threshold
    center_row_is_black_deriv = np.diff(center_row_is_black.astype(int))

    # first and last 1 in the derivative
    center_row_black_start = np.where(center_row_is_black_deriv == 1)[0][0]
    center_row_black_end = np.where(center_row_is_black_deriv == -1)[0][-1]
    radius = (center_row_black_end - center_row_black_start) / 2
    return radius


@attrs.define()
class NormalizeTransformer(TransformerBase):
    """Normalize the coordinates to [-1, 1]."""

    center: tuple[float, float] | None = None
    """The center of the image. If None, the center is the center of the image."""
    scale: tuple[float, float] | Literal["min", "max"] | None = None
    """The scale of the image. If "min" or None, the scale is the minimum of the width and height.
    If "max", the scale is the maximum of the width and height."""

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        center = self.center or (x.shape[1] / 2, x.shape[0] / 2)
        scale = (
            min(x.shape[1], x.shape[0])
            if self.scale in ["min", None]
            else max(x.shape[1], x.shape[0]) if self.scale == "max" else self.scale
        )
        x = (x - center[0]) / scale * 2
        y = (y - center[1]) / scale * 2
        return x, y

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        center = self.center or (x.shape[1] / 2, x.shape[0] / 2)
        scale = (
            min(x.shape[1], x.shape[0])
            if self.scale in ["min", None]
            else max(x.shape[1], x.shape[0]) if self.scale == "max" else self.scale
        )
        x = x * scale[0] + center[0]
        y = y * scale[1] + center[1]
        return x, y


# @attrs.define()
# class AutoDetectRadiusNormalizeTransformer(NormalizeTransformer):
#     def fit(self, image: NDArray, **kwargs: Any) -> None:
#         radius = get_radius(image)
#         self.center = (image.shape[1] // 2, image.shape[0] // 2)
#         self.scale = (radius, radius)


@attrs.define()
class DenormalizeTransformer(TransformerBase):
    """Denormalize the coordinates from [-1, 1] to the original image size."""

    scale: tuple[float, float]
    """The scale of the image. Recommended to be the half of the width and height of the result image."""
    center: tuple[float, float]
    """The center of the image. Recommended to be the center of the result image."""

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        scale = self.scale
        center = self.center
        x = x * scale[0] + center[0]
        y = y * scale[1] + center[1]
        return x, y

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        scale = self.scale
        center = self.center
        x = (x - center[0]) / scale[0]
        y = (y - center[1]) / scale[1]
        return x, y


@attrs.define()
class PolarRollTransformer(TransformerBase):
    """Transform using polar coordinates."""

    @abstractmethod
    def transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        """
        Transform using polar coordinates.

        Parameters
        ----------
        theta : NDArray
            The distance or angle from the center (front-facing direction)
        roll : NDArray
            The angle around the center (front-facing direction)
        **kwargs : Any
            Any additional keyword arguments.

        Returns
        -------
        tuple[NDArray, NDArray]
            theta and roll after transformation.

        """
        pass

    @abstractmethod
    def inverse_transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        """
        Inverse transform using polar coordinates.

        Parameters
        ----------
        theta : NDArray
            The distance or angle from the center (front-facing direction)
        roll : NDArray
            The angle around the center (front-facing direction)
        **kwargs : Any
            Any additional keyword arguments.

        Returns
        -------
        tuple[NDArray, NDArray]
            theta and roll after transformation.

        """
        pass

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        theta = np.sqrt(x**2 + y**2)
        roll = np.arctan2(y, x)
        theta, roll = self.transform_polar(theta, roll, **kwargs)
        x = theta * np.cos(roll)
        y = theta * np.sin(roll)
        return x, y

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        theta = np.sqrt(x**2 + y**2)
        roll = np.arctan2(y, x)
        theta, roll = self.inverse_transform_polar(theta, roll, **kwargs)
        x = theta * np.cos(roll)
        y = theta * np.sin(roll)
        return x, y


@attrs.define()
class RectilinearDecoder(PolarRollTransformer):
    """Encodes rectilinear image."""

    # https://en.wikipedia.org/wiki/Image_sensor_format
    focal_length: float
    """The focal length of the lens in mm."""
    sensor_width: (
        Literal["35mm", "APS-H", "APS-C", "APS-C-Canon", "Foveon", "MFT"] | str | float
    ) = "35mm"
    """The sensor width of the camera in mm if float,
    or in inches if str, or a standard sensor width if str."""

    @property
    def sensor_width_mm(self) -> float:
        """Sensor width in mm."""
        if self.sensor_width in ["35mm", "APS-C", "1/2.3"]:
            warnings.warn(
                "Sensor size may vary by about 0.2 mm depending on the camera model. "
                "To get very accurate results, consider setting the sensor width in mm manually.",
                UserWarning,
                stacklevel=2,
            )
        # https://en.wikipedia.org/wiki/Image_sensor_format#Table_of_sensor_formats_and_sizes
        known_widths = {
            "35mm": 36.0,  # ~ 35.8mm
            "APS-H": 27.90,
            "APS-C": 23.6,  # ~ 23.7mm
            "APS-C-Canon": 22.30,
            "MFT": 17.30,
            "1": 13.20,
            "1/1.12": 11.43,
            "1/1.2": 10.67,
            "1/1.33": 9.6,
            "1/1.6": 8.08,
            "1/1.7": 7.60,
            "1/1.8": 7.18,
            "1/2": 6.40,
            "1/2.3": 6.17,
        }
        if isinstance(self.sensor_width, str):
            return known_widths[self.sensor_width]
        return self.sensor_width

    @property
    def factor(self) -> float:
        """Zoom factor applied after tan."""
        return 2 * self.focal_length / self.sensor_width_mm

    def transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        return np.tan(theta) * self.factor, roll

    def inverse_transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        # fov = 2 arctan sensor_width / (2 * focal_length)
        return np.arctan(theta / self.factor), roll


@attrs.define()
class FisheyeEncoder(PolarRollTransformer):
    """Encodes fisheye image."""

    mapping_type: Literal[
        "rectilinear", "stereographic", "equidistant", "equisolid", "orthographic"
    ]
    """The mapping type of the fisheye image."""

    def transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        """[-1, 1] -> [-pi/2, pi/2]."""
        if self.mapping_type == "rectilinear":
            return np.arctan(theta), roll
        elif self.mapping_type == "stereographic":
            return 2 * np.arctan(theta), roll
        elif self.mapping_type == "equidistant":
            return theta * (np.pi / 2), roll
        elif self.mapping_type == "equisolid":
            return 2 * np.arcsin(theta / np.sqrt(2)), roll
        elif self.mapping_type == "orthographic":
            return np.arcsin(theta), roll
        else:
            raise ValueError(
                f"Unknown mapping type: {self.mapping_type}, "
                "should be one of 'rectilinear', 'stereographic', 'equidistant', 'equisolid', 'orthographic'."
            )

    def inverse_transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        """[-pi/2, pi/2] -> [-1, 1]."""
        if self.mapping_type == "rectilinear":
            return np.tan(theta), roll
        elif self.mapping_type == "stereographic":
            return 2 * np.tan(theta / 2), roll
        elif self.mapping_type == "equidistant":
            return theta / (np.pi / 2), roll
        elif self.mapping_type == "equisolid":
            return np.sqrt(2) * np.sin(theta / 2), roll
        elif self.mapping_type == "orthographic":
            return np.sin(theta), roll
        else:
            raise ValueError(
                f"Unknown mapping type: {self.mapping_type}, "
                "should be one of 'rectilinear', 'stereographic', 'equidistant', 'equisolid', 'orthographic'."
            )


@attrs.define()
class InverseTransformer(TransformerBase, Generic[T]):
    """Transformer that calls inverse_transform() in transform() and vice versa."""

    transformer: T
    """The transformer to be inverted."""

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        return self.transformer.inverse_transform(x, y, **kwargs)

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        return self.transformer.transform(x, y, **kwargs)


def FisheyeDecoder(
    mapping_type: Literal[
        "rectilinear", "stereographic", "equidistant", "equisolid", "orthographic"
    ]
) -> InverseTransformer[FisheyeEncoder]:
    """
    Decodes fisheye image.

    Parameters
    ----------
    mapping_type : Literal['rectilinear', 'stereographic', 'equidistant', 'equisolid', 'orthographic']
        The mapping type of the fisheye image.

    Returns
    -------
    InverseTransformer
        The fisheye decoder.

    """
    return InverseTransformer(FisheyeEncoder(mapping_type))


@attrs.define()
class PolynomialScaler(PolarRollTransformer):
    """Scale the polar coordinates using polynomial."""

    coefs_reverse: Sequence[float] = [0, 1]
    """The coefficients of the polynomial in reverse order.
    [0, 1] means y = 0 + 1 * x."""

    def transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        return np.polyval(np.flip(self.coefs_reverse), theta), roll

    def inverse_transform_polar(
        self, theta: NDArray, roll: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        raise NotImplementedError(
            "PolynomialScaler does not support inverse transform."
        )


@attrs.define()
class ZoomTransformer(TransformerBase):
    """Zoom the image."""

    scale: float
    """The zoom scale."""

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        x = x / self.scale
        y = y / self.scale
        return x, y

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        x = x * self.scale
        y = y * self.scale
        return x, y


def equidistant_to_3d(x: NDArray, y: NDArray) -> NDArray:
    """
    Convert 2D coordinates to 3D unit vector.

    z axis is forward, x axis is right, y axis is up.

    Parameters
    ----------
    x : NDArray
        The x coordinate in equidistant fisheye format.
    y : NDArray
        The y coordinate in equidistant fisheye format.

    Returns
    -------
    NDArray
        The 3D unit vector.

    """
    phi = np.arctan2(x, y)
    theta = np.sqrt(x**2 + y**2)
    v = np.stack(
        [np.sin(theta) * np.sin(phi), np.sin(theta) * np.cos(phi), np.cos(theta)],
        axis=-1,
    )
    return v


def equidistant_from_3d(v: NDArray) -> tuple[NDArray, NDArray]:
    """
    Convert 3D unit vector to 2D coordinates.

    Parameters
    ----------
    v : NDArray
        The 3D unit vector.

    Returns
    -------
    tuple[NDArray, NDArray]
        The x and y coordinates in equidistant fisheye format.

    """
    theta = np.arccos(v[..., 2])
    phi = np.arctan2(v[..., 0], v[..., 1])
    x = theta * np.sin(phi)
    y = theta * np.cos(phi)
    return x, y


@attrs.define()
class EquirectangularEncoder(TransformerBase):
    """Encodes equirectangular image."""

    is_latitude_y: bool = True
    """Whether latitude is encoded in y axis."""

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        # latitude: 日本語で緯度, phi
        # longitude: 日本語で経度, theta
        if self.is_latitude_y:
            theta_lat = y * (np.pi / 2)
            phi_lon = x * (np.pi / 2)
            v = np.stack(
                [
                    np.cos(theta_lat) * np.sin(phi_lon),
                    np.sin(theta_lat),
                    np.cos(theta_lat) * np.cos(phi_lon),
                ],
                axis=-1,
            )
        else:
            theta_lat = x * (np.pi / 2)
            phi_lon = y * (np.pi / 2)
            v = np.stack(
                [
                    np.sin(theta_lat),
                    np.cos(theta_lat) * np.sin(phi_lon),
                    np.cos(theta_lat) * np.cos(phi_lon),
                ],
                axis=-1,
            )

        return equidistant_from_3d(v)

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        v = equidistant_to_3d(x, y)
        if self.is_latitude_y:
            theta_lat = np.arcsin(v[..., 1])
            phi_lon = np.arctan2(v[..., 0], v[..., 2])
            x = phi_lon / (np.pi / 2)
            y = theta_lat / (np.pi / 2)
        else:
            theta_lat = np.arcsin(v[..., 0])
            phi_lon = np.arctan2(v[..., 1], v[..., 2])
            x = theta_lat / (np.pi / 2)
            y = phi_lon / (np.pi / 2)
        return x, y


def EquirectangularDecoder(
    is_latitude_y: bool = True,
) -> InverseTransformer[EquirectangularEncoder]:
    """
    Decodes equirectangular image.

    Parameters
    ----------
    is_latitude_y : bool, optional
        Whether latitude is encoded in y axis, by default True

    Returns
    -------
    InverseTransformer[EquirectangularEncoder]
        The equirectangular decoder.

    """
    return InverseTransformer(EquirectangularEncoder(is_latitude_y))


@attrs.define()
class Euclidean3DTransformer(TransformerBase):
    """Transform as 3D unit vector."""

    @abstractmethod
    def transform_v(self, v: NDArray) -> NDArray:
        """
        Transform 3D unit vector.

        Parameters
        ----------
        v : NDArray
            The 3D unit vector.
            z axis is forward, x axis is right, y axis is up.

        Returns
        -------
        NDArray
            The transformed 3D unit vector.
            z axis is forward, x axis is right, y axis is up.

        """
        pass

    @abstractmethod
    def inverse_transform_v(self, v: NDArray) -> NDArray:
        """
        Inverse transform 3D unit vector.

        Parameters
        ----------
        v : NDArray
            The 3D unit vector.
            z axis is forward, x axis is right, y axis is up.

        Returns
        -------
        NDArray
            The inverse transformed 3D unit vector.
            z axis is forward, x axis is right, y axis is up.

        """
        pass

    def transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        v = equidistant_to_3d(x, y)
        v = self.transform_v(v)
        x, y = equidistant_from_3d(v)
        return x, y

    def inverse_transform(
        self, x: NDArray, y: NDArray, **kwargs: Any
    ) -> tuple[NDArray, NDArray]:
        v = equidistant_to_3d(x, y)
        v = self.transform_v(v)
        x, y = equidistant_from_3d(v)
        return x, y


@attrs.define()
class Euclidean3DRotator(Euclidean3DTransformer):
    """Rotate as 3D unit vector."""

    rotation: quaternion
    """The rotation quaternion."""

    def transform_v(self, v: NDArray) -> NDArray:
        return rotate_vectors(self.rotation, v)

    def inverse_transform_v(self, v: NDArray) -> NDArray:
        return rotate_vectors(self.rotation.inverse(), v)



from datetime import datetime, timezone
from enum import auto
from hashlib import sha256
from logging import DEBUG, INFO, basicConfig, getLogger
from pathlib import Path
from typing import Any, Sequence

import cv2 as cv
import typer
from quaternion import *  # noqa
from rich.logging import RichHandler
from strenum import StrEnum
from typing_extensions import Annotated

from vr180_convert.transformer import *  # noqa
from vr180_convert.transformer import (
    EquirectangularEncoder,
    Euclidean3DRotator,
    FisheyeDecoder,
    MultiTransformer,
    TransformerBase,
)

from .remapper import apply, apply_lr, match_lr

LOG = getLogger(__name__)
DEFAULT_EXTENSION = "png"

app = typer.Typer()


@app.callback()
def _main(verbose: bool = typer.Option(False, "--verbose", "-v")) -> None:
    level = INFO
    if verbose:
        level = DEBUG
    basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )


class _InterpolationFlags(StrEnum):
    """Interpolation flags enum for typer."""

    INTER_NEAREST = auto()
    INTER_LINEAR = auto()
    INTER_CUBIC = auto()
    INTER_AREA = auto()
    INTER_LANCZOS4 = auto()
    INTER_MAX = auto()
    WARP_FILL_OUTLIERS = auto()
    WARP_INVERSE_MAP = auto()


class _BorderTypes(StrEnum):
    """Border types enum for typer."""

    BORDER_CONSTANT = auto()
    BORDER_REPLICATE = auto()
    BORDER_REFLECT = auto()
    BORDER_WRAP = auto()
    BORDER_REFLECT_101 = auto()
    BORDER_TRANSPARENT = auto()
    BORDER_ISOLATED = auto()


def _get_position_gui(image_paths: Sequence[Path]) -> list[tuple[int, int]]:
    """Get the position of the GUI window."""
    window_name = "Select position"
    cv.namedWindow(window_name, cv.WND_PROP_FULLSCREEN)
    cv.setWindowProperty(window_name, cv.WND_PROP_FULLSCREEN, cv.WINDOW_FULLSCREEN)

    i = 0
    res = []

    def on_mouse(event: int, x: int, y: int, flags: int, param: Any) -> None:
        nonlocal res, i
        if event == cv.EVENT_LBUTTONDOWN:
            res.append((x, y))
            LOG.info(f"Position {i}: ({x}, {y})")

    cv.setMouseCallback(window_name, on_mouse)
    cv.imshow(window_name, cv.imread(image_paths[i].as_posix()))
    while True:
        cv.waitKey(10)
        if len(res) == i + 1:
            i += 1
            if i == len(image_paths):
                break
            cv.imshow(window_name, cv.imread(image_paths[i].as_posix()))
    cv.destroyAllWindows()
    return res


@app.command()
def lr(
    left_path: Annotated[Path, typer.Argument(help="Left image path")],
    right_path: Annotated[Path, typer.Argument(help="Right image path")],
    transformer: Annotated[
        str, typer.Option(help="Transformer Python code (to be `eval()`ed)")
    ] = "",
    out_path: Annotated[
        Path,
        typer.Option(
            help="Output image path, defaults to left_path.with_suffix('.out.jpg')"
        ),
    ] = Path(""),
    size: Annotated[
        str, typer.Option(help="Output image size, defaults to 4096x4096")
    ] = "4096x4096",
    interpolation: Annotated[
        _InterpolationFlags,
        typer.Option(help="Interpolation method, defaults to lanczos4"),
    ] = _InterpolationFlags.INTER_LANCZOS4,  # type: ignore
    border_mode: Annotated[
        _BorderTypes, typer.Option(help="Border mode, defaults to constant")
    ] = _BorderTypes.BORDER_CONSTANT,  # type: ignore
    border_value: int = 0,
    radius: Annotated[
        str, typer.Option(help="Radius of the fisheye image, defaults to 'auto'")
    ] = "auto",
    merge: Annotated[bool, typer.Option(help="Export as an anaglyph")] = False,
    autosearch_timestamp_calib_r_earlier_l: Annotated[
        float,
        typer.Option(
            "--autosearch-timestamp-calib-r-earlier-l",
            "-ac",
            help="Autosearch timestamp calibration "
            "(right timestamp -= autosearch_timestamp_calib_r_earlier_l) (in seconds)",
        ),
    ] = 0.0,
    swap: Annotated[
        bool,
        typer.Option(help="Swap left and right images as well as transformer, etc."),
    ] = False,
    name_unique: Annotated[bool, typer.Option(help="Make output name unique")] = False,
    automatch: Annotated[
        str,
        typer.Option(
            help='Calibrate rotation. e.g. "0,0;0,0;1,1;1,1". If "gui", use GUI'
        ),
    ] = "",
) -> None:
    """Remap a pair of fisheye images to a pair of SBS equirectangular images."""
    if swap:
        left_path, right_path = right_path, left_path
        autosearch_timestamp_calib_r_earlier_l = -autosearch_timestamp_calib_r_earlier_l

    # find closest time-matched images
    if left_path.is_dir() and not right_path.is_dir():
        # find closest time-matched right image
        # sort with time
        right_time = right_path.stat().st_mtime
        left_path_candidates = sorted(
            left_path.rglob("*"),
            key=lambda p: abs(
                p.stat().st_mtime - right_time + autosearch_timestamp_calib_r_earlier_l
            ),
            reverse=False,
        )
        left_path_candidates = [
            p
            for p in left_path_candidates
            if (p != right_path) and (p.suffix == right_path.suffix)
        ]
        if len(left_path_candidates) == 0:
            raise ValueError("No time-matched left image found")
        left_path = left_path_candidates[0]
    elif not left_path.is_dir() and right_path.is_dir():
        # find closest time-matched left image
        # sort with time
        left_time = left_path.stat().st_mtime
        right_path_candidates = sorted(
            right_path.rglob("*"),
            key=lambda p: abs(
                p.stat().st_mtime - left_time - autosearch_timestamp_calib_r_earlier_l
            ),
            reverse=False,
        )
        right_path_candidates = [
            p
            for p in right_path_candidates
            if (p != left_path) and (p.suffix == left_path.suffix)
        ]
        if len(right_path_candidates) == 0:
            raise ValueError("No time-matched right image found")
        right_path = right_path_candidates[0]
    elif left_path.is_dir() and right_path.is_dir():
        raise ValueError("Both left and right paths must not be directories")

    LOG.info(
        f"L: {left_path}"
        f"@{datetime.fromtimestamp(left_path.stat().st_mtime, timezone.utc)}, "
        f"R: {right_path}"
        f"@{datetime.fromtimestamp(right_path.stat().st_mtime, timezone.utc)}"
    )

    # evaluate automatch
    transformer_: TransformerBase | tuple[TransformerBase, TransformerBase]
    # evaluate transformer
    if transformer == "":
        transformer_ = EquirectangularEncoder() * FisheyeDecoder("equidistant")
    else:
        transformer_ = eval(transformer)  # noqa
    if automatch != "":
        if not isinstance(transformer_, MultiTransformer):
            raise ValueError("Automatch requires MultiTransformer")

        transformer_is_encoder = [
            x.__class__.__name__.endswith("Encoder") for x in transformer_.transformers
        ]
        transformer_first_encoder_index = transformer_is_encoder.index(True)
        transformer_until_encoder = MultiTransformer(
            transformer_.transformers[: transformer_first_encoder_index + 1]
        )
        transformer_after_encoder = MultiTransformer(
            transformer_.transformers[transformer_first_encoder_index + 1 :]
        )
        LOG.debug(f"{transformer_until_encoder=}, {transformer_after_encoder=}")

        if automatch == "gui":
            automatch_ = _get_position_gui(
                [left_path, right_path, left_path, right_path]
            )
            LOG.info(
                f"Automatched position: {';'.join([','.join(map(str, p)) for p in automatch_])}"
            )
        else:
            automatch_ = [tuple(chunk.split(",")) for chunk in automatch.split(";")]  # type: ignore

        q = match_lr(
            FisheyeDecoder("equidistant"),
            # odd
            automatch_[1::2],
            # even
            automatch_[::2],
            radius=float(radius) if radius not in ["auto", "max"] else radius,  # type: ignore
            in_paths=[left_path, right_path],
        )
        LOG.info(f"Automatched quaternion: {q}")
        transformer_ = (
            transformer_until_encoder
            * Euclidean3DRotator(q)
            * transformer_after_encoder,
            transformer_,
        )
        LOG.info(f"Automatched transformer: {transformer_}")

    # if swap:
    #     if isinstance(transformer_, tuple) and automatch == "":
    #         transformer_ = transformer_[1], transformer_[0]

    # apply transformer
    name_unique_content = (
        (
            "-"
            + sha256(
                "".join(
                    [
                        transformer,
                        size,
                        interpolation,
                        border_mode,
                        str(border_value),
                        radius,
                        str(merge),
                        str(autosearch_timestamp_calib_r_earlier_l),
                        str(swap),
                    ]
                ).encode("utf-8")
            ).hexdigest()[:8]
        )
        if name_unique
        else ""
    )
    filename_default = (
        f"{Path(left_path).stem}-"
        + f"{Path(right_path).stem}{name_unique_content}.{DEFAULT_EXTENSION}"
    )
    apply_lr(
        transformer=transformer_,
        left_path=left_path,
        right_path=right_path,
        out_path=(
            Path(left_path).parent / filename_default
            if out_path == Path("")
            else out_path / filename_default if out_path.is_dir() else out_path
        ),
        radius=float(radius) if radius not in ["auto", "max"] else radius,  # type: ignore
        size_output=tuple(map(int, size.split("x"))),  # type: ignore
        interpolation=getattr(cv, interpolation.upper()),
        boarder_mode=getattr(cv, border_mode.upper()),
        boarder_value=border_value,
        merge=merge,
    )


@app.command()
def s(
    in_paths: Annotated[list[Path], typer.Argument(help="Image paths")],
    transformer: Annotated[
        str, typer.Option(help="Transformer Python code (to be `eval()`ed)")
    ] = "",
    out_path: Annotated[
        Path,
        typer.Option(
            help="Output image path, defaults to left_path.with_suffix('.out.jpg')"
        ),
    ] = Path(""),
    size: Annotated[
        str, typer.Option(help="Output image size, defaults to 4096x4096")
    ] = "4096x4096",
    interpolation: Annotated[
        _InterpolationFlags,
        typer.Option(help="Interpolation method, defaults to lanczos4"),
    ] = _InterpolationFlags.INTER_LANCZOS4,  # type: ignore
    boarder_mode: Annotated[
        _BorderTypes, typer.Option(help="Border mode, defaults to constant")
    ] = _BorderTypes.BORDER_CONSTANT,  # type: ignore
    boarder_value: int = 0,
    radius: Annotated[
        str, typer.Option(help="Radius of the fisheye image, defaults to 'auto'")
    ] = "auto",
) -> None:
    """Remap fisheye images to SBS equirectangular images."""
    if transformer == "":
        transformer_ = EquirectangularEncoder() * FisheyeDecoder("equidistant")
    else:
        transformer_ = eval(transformer)  # noqa

    if out_path == Path(""):
        out_paths = [p.with_suffix(f".out.{DEFAULT_EXTENSION}") for p in in_paths]
    elif out_path.is_dir():
        out_paths = [out_path / p.name for p in in_paths]
    else:
        if len(in_paths) > 1:
            raise ValueError(
                "Output path must be a directory when multiple input paths are provided"
            )
        out_paths = [out_path for p in in_paths]

    apply(
        transformer=transformer_,
        in_paths=in_paths,
        out_paths=out_paths,
        radius=float(radius) if radius not in ["auto", "max"] else radius,  # type: ignore
        size_output=tuple(map(int, size.split("x"))),  # type: ignore
        interpolation=getattr(cv, interpolation.upper()),
        boarder_mode=getattr(cv, boarder_mode.upper()),
        boarder_value=boarder_value,
    )



from __future__ import annotations

from logging import getLogger
from pathlib import Path
from typing import Literal, Sequence

import cv2 as cv
import numpy as np
from numpy.typing import NDArray
from quaternion import as_quat_array, quaternion

from .transformer import (
    DenormalizeTransformer,
    NormalizeTransformer,
    TransformerBase,
    equidistant_to_3d,
    get_radius,
)

LOG = getLogger(__name__)


def get_map(
    transformer: TransformerBase,
    *,
    radius: float,
    size_input: tuple[int, int],
    size_output: tuple[int, int] = (2048, 2048),
) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
    """
    Generate remap map.

    Parameters
    ----------
    transformer : TransformerBase
        Transformer to be applied.
    radius : float
        Radius of the fisheye image.
    size_input : tuple[int, int]
        Size of the input image.
    size_output : tuple[int, int], optional
        Size of the output image, by default (2048, 2048)

    Returns
    -------
    tuple[NDArray[np.float32], NDArray[np.float32]]
        xmap and ymap.

    """
    xmap, ymap = np.meshgrid(np.arange(size_output[0]), np.arange(size_output[1]))
    xmap, ymap = (
        NormalizeTransformer()
        * transformer
        * DenormalizeTransformer(
            scale=(radius, radius), center=(size_input[1] // 2, size_input[0] // 2)
        )
    ).transform(xmap, ymap)
    xmap, ymap = xmap.astype(np.float32), ymap.astype(np.float32)
    return xmap, ymap


def get_radius_smart(
    radius: float | Literal["auto", "max"],
    images: Sequence[NDArray],
) -> float:
    """
    Get radius smartly.

    Parameters
    ----------
    radius : float | Literal[&quot;auto&quot;, &quot;max&quot;]
        The strategy to get the radius.
    images : Sequence[NDArray]
        Images to be processed.

    Returns
    -------
    float
        The radius.

    """
    if radius == "auto":
        radius_candidates = [get_radius(image) for image in images]
        radius_ = max(radius_candidates)
    elif radius == "max":
        radius_ = min(images[0].shape[0] / 2, images[0].shape[1] / 2)
    else:
        radius_ = radius
    LOG.info(f"Radius: {radius_}, strategy: {radius}, image shape: {images[0].shape}")
    return radius_


def rotation_match(
    points_to_be_rotated: NDArray,
    points: NDArray,
) -> quaternion:
    """
    Match the rotation of two sets of 3d points.

    Parameters
    ----------
    points_to_be_rotated : _type_
        Array of shape (..., 3)
    points : NDArray
        Array of shape (..., 3)

    Returns
    -------
    quaternion
        quaternion that minimizes the distance between the rotated
        `points_to_be_rotated` and `points`.

    References
    ----------
    https://lisyarus.github.io/blog/posts/3d-shape-matching-with-quaternions.html

    """
    # 3d point matching
    # https://lisyarus.github.io/blog/posts/3d-shape-matching-with-quaternions.html
    # E := ||Ra_k - b_k||^2, Ra := qaq^{-1}
    # E = ||qa_kq^{-1} - b_k||^2 = ||qa_k - b_kq||^2

    # extend to 4d
    a = np.concatenate(
        [points_to_be_rotated, np.zeros_like(points_to_be_rotated[..., :1])], axis=1
    )
    ax, ay, az, aw = a[..., 0], a[..., 1], a[..., 2], a[..., 3]
    b = np.concatenate([points, np.zeros_like(points[..., :1])], axis=1)
    bx, by, bz, bw = b[..., 0], b[..., 1], b[..., 2], b[..., 3]

    right_mult_matrix = np.array(
        [[aw, -az, ay, -ax], [az, aw, -ax, -ay], [-ay, ax, aw, -az], [ax, ay, az, aw]]
    )
    left_mult_matrix = np.array(
        [[bw, bz, -by, -bx], [-bz, bw, bx, -by], [by, -bx, bw, -bz], [bx, by, bz, bw]]
    )
    S = right_mult_matrix - left_mult_matrix
    B = np.einsum("jik,jlk->il", S, S)
    eigenvalues, eigenvectors = np.linalg.eig(B)
    q = eigenvectors[:, np.argmin(eigenvalues)]
    return as_quat_array([q[..., 3], q[..., 0], q[..., 1], q[..., 2]])


def match_lr(
    decoder: TransformerBase,
    points_l: Sequence[tuple[float, float]],
    points_r: Sequence[tuple[float, float]],
    in_paths: Sequence[Path | str],
    *,
    radius: float | Literal["auto", "max"] = "auto",
) -> quaternion:
    """
    Get the quaternion that minimizes the distance between the rotated points.

    Parameters
    ----------
    decoder : TransformerBase
        Transformer to be applied. Must implement inverse_transform().
    points_l : tuple[tuple[float, float], ...]
        The points in the left image.
    points_r : tuple[tuple[float, float], ...]
        The points in the right image.
    in_paths : Sequence[Path  |  str]
        Input image paths.
    radius : float | Literal[&quot;auto&quot;, &quot;max&quot], optional
        The strategy to get the radius.

    Returns
    -------
    quaternion
        The quaternion to be applied to the left points.

    Raises
    ------
    ValueError
        If the number of points is not the same.

    """
    if len(points_l) != len(points_r):
        raise ValueError("The number of points must be the same.")
    points_ = np.concatenate([points_l, points_r], axis=0)
    images = [cv.imread(Path(from_path).as_posix()) for from_path in in_paths]
    xmap, ymap = points_[:, 0], points_[:, 1]
    xmap, ymap = xmap.astype(np.float32), ymap.astype(np.float32)
    xmap, ymap = (
        decoder
        * DenormalizeTransformer(
            scale=(radius, radius),
            center=(images[0].shape[1] // 2, images[0].shape[0] // 2),
        )
    ).inverse_transform(xmap, ymap)
    v = equidistant_to_3d(xmap, ymap)
    vl, vr = v[: len(points_l)], v[len(points_l) :]
    return rotation_match(vl, vr)


def apply(
    transformer: TransformerBase,
    *,
    in_paths: Sequence[Path | str | NDArray] | Path | str | NDArray,
    out_paths: Sequence[Path | str] | None | Path | str = None,
    size_output: tuple[int, int] = (2048, 2048),
    interpolation: int = cv.INTER_LANCZOS4,
    boarder_mode: int = cv.BORDER_CONSTANT,
    boarder_value: int | tuple[int, int, int] = 0,
    radius: float | Literal["auto", "max"] = "auto",
) -> Sequence[NDArray[np.uint8]]:
    """
    Apply transformer to images.

    Parameters
    ----------
    transformer : TransformerBase
        Transformer to be applied.
    in_paths : Sequence[Path  |  str] | Path | str
        Input image paths.
    out_paths : Sequence[Path  |  str] | None | Path | str, optional
        Output image paths, by default None
        If None, do not save the images.
    size_output : tuple[int, int], optional
        Size of the output image, by default (2048, 2048)
    interpolation : int, optional
        Interpolation method for opencv, by default cv.INTER_LANCZOS4
    boarder_mode : int, optional
        Boarder mode for opencv, by default cv.BORDER_CONSTANT
    boarder_value : int | tuple[int, int, int], optional
        Boarder value for opencv, by default 0
    radius : float | Literal[&quot;auto&quot;, &quot;max&quot;], optional
        Radius of the fisheye image, by default &quot;auto&quot;

    Returns
    -------
    Sequence[NDArray[np.uint8]]
        Images after transformation.

    """
    # note that str is Sequence
    in_paths_ = (
        [in_paths] if isinstance(in_paths, (str, Path, np.ndarray)) else in_paths
    )
    out_paths_ = [out_paths] if isinstance(out_paths, (str, Path)) else out_paths
    del in_paths, out_paths

    images = [
        (
            cv.imread(Path(from_path).as_posix())
            if isinstance(from_path, (str, Path))
            else from_path
        )
        for from_path in in_paths_
    ]
    radius_ = get_radius_smart(radius, images)

    xmap, ymap = get_map(
        radius=radius_,
        transformer=transformer,
        size_output=size_output,
        size_input=(images[0].shape[0], images[0].shape[1]),
    )

    images = [
        cv.remap(
            img,
            xmap,
            ymap,
            interpolation=interpolation,
            borderMode=boarder_mode,
            borderValue=boarder_value,
        )
        for img in images
    ]

    if out_paths_ is not None:
        for to_path, image in zip(out_paths_, images):
            cv.imwrite(Path(to_path).as_posix(), image)
    return images


def apply_lr(
    transformer: TransformerBase | tuple[TransformerBase, TransformerBase],
    *,
    left_path: Path | str | NDArray,
    right_path: Path | str | NDArray,
    out_path: Path | str,
    size_output: tuple[int, int] = (2048, 2048),
    interpolation: int = cv.INTER_LANCZOS4,
    boarder_mode: int = cv.BORDER_CONSTANT,
    boarder_value: int | tuple[int, int, int] = 0,
    radius: float | Literal["auto", "max"] = "auto",
    merge: bool = False,
) -> None:
    """
    Apply transformer to a pair of images.

    Parameters
    ----------
    transformer : TransformerBase
        Transformer to be applied.
    left_path : Path | str
        Left image path.
    right_path : Path | str
        Right image path.
    in_paths : Sequence[Path  |  str] | Path | str
        Input image paths.
    out_path : Path | str
        Output image path.
    size_output : tuple[int, int], optional
        Size of the output image, by default (2048, 2048)
    interpolation : int, optional
        Interpolation method for opencv, by default cv.INTER_LANCZOS4
    boarder_mode : int, optional
        Boarder mode for opencv, by default cv.BORDER_CONSTANT
    boarder_value : int | tuple[int, int, int], optional
        Boarder value for opencv, by default 0
    radius : float | Literal[&quot;auto&quot;, &quot;max&quot;], optional
        Radius of the fisheye image, by default &quot;auto&quot;
    merge : bool, optional
        Whether to merge the images mainly for calibration, by default False

    """
    if (
        isinstance(left_path, (str, Path))
        and isinstance(right_path, (str, Path))
        and left_path == right_path
    ):
        image = cv.imread(Path(left_path).as_posix())
        # divide left and right
        left_path = image[:, : image.shape[1] // 2]
        right_path = image[:, image.shape[1] // 2 :]

    images: Sequence[NDArray[np.uint8]]

    if isinstance(transformer, tuple):
        images = [
            apply(
                in_paths=in_path,
                out_paths=None,
                transformer=transformer,
                size_output=size_output,
                interpolation=interpolation,
                boarder_mode=boarder_mode,
                boarder_value=boarder_value,
                radius=radius,
            )[0]
            for transformer, in_path in zip(transformer, [left_path, right_path])
        ]
    else:
        images = apply(
            in_paths=[left_path, right_path],
            out_paths=None,
            transformer=transformer,
            size_output=size_output,
            interpolation=interpolation,
            boarder_mode=boarder_mode,
            boarder_value=boarder_value,
            radius=radius,
        )
    if merge:
        # https://en.wikipedia.org/wiki/Anaglyph_3D
        # 3d glass -> L: red filter, R: blue filter
        # L: blue layer, R: red layer
        colors = [(0, 128, 255), (255, 128, 0)]
        combine = np.mean(images[0], axis=-1)[..., None] * np.array(colors[0]).reshape(
            [1] * (images[0].ndim - 1) + [3]
        ) + (
            np.mean(images[1], axis=-1)[..., None]
            * np.array(colors[1]).reshape([1] * (images[1].ndim - 1) + [3])
        )
        combine /= 255
        cv.putText(
            combine,
            "L",
            (0, len(combine[1]) // 10),
            cv.FONT_HERSHEY_SIMPLEX,
            len(combine) // 1000,
            colors[0],
            2,
            cv.LINE_AA,
        )
        cv.putText(
            combine,
            "R",
            (len(combine[1]) // 2, len(combine[0]) // 10),
            cv.FONT_HERSHEY_SIMPLEX,
            len(combine) // 1000,
            colors[1],
            2,
            cv.LINE_AA,
        )
    else:
        combine = np.concatenate(images, axis=1)
    cv.imwrite(Path(out_path).as_posix(), combine)
    LOG.info(f"Saved to {Path(out_path).absolute()}")
